from diamondback.transforms.ComplexTransform import ComplexTransform
from diamondback.transforms.FourierTransform import FourierTransform
from diamondback.transforms.PowerSpectrumTransform import PowerSpectrumTransform
from diamondback.transforms.WaveletTransform import WaveletTransform
from diamondback.transforms.ZTransform import ZTransform



""" **Description**

        State interface.

    **Example**

        ::

            from diamondback.commons.Serial import Serial
            from diamondback.interfaces.IState import IState

            class Test( IState ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.state = ''

            test = Test( )

            test.state = Serial.encode( { 'a' : 0.0, 'b' : 1.0 }, compress = False )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-12.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual


class IState( IEqual ) :

    """ State interface.
    """

    @property
    def state( self ) :

        """ State ( any ).
        """

        return self._state

    @state.setter
    def state( self, state : any ) :

        self._state = state

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other object ( object ).

            Returns :

                equality - Equality condition ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self.state == other.state ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._state = [ ]



from diamondback.interfaces.IA import IA
from diamondback.interfaces.IB import IB
from diamondback.interfaces.ICache import ICache
from diamondback.interfaces.IClear import IClear
from diamondback.interfaces.ICompress import ICompress
from diamondback.interfaces.IConfigure import IConfigure
from diamondback.interfaces.IConnect import IConnect
from diamondback.interfaces.IData import IData
from diamondback.interfaces.IDate import IDate
from diamondback.interfaces.IDispose import IDispose
from diamondback.interfaces.IDuration import IDuration
from diamondback.interfaces.IEmulate import IEmulate
from diamondback.interfaces.IEncoding import IEncoding
from diamondback.interfaces.IEqual import IEqual
from diamondback.interfaces.IFrequency import IFrequency
from diamondback.interfaces.IIdentity import IIdentity
from diamondback.interfaces.IInterval import IInterval
from diamondback.interfaces.ILatency import ILatency
from diamondback.interfaces.IPath import IPath
from diamondback.interfaces.IPeriod import IPeriod
from diamondback.interfaces.IPhase import IPhase
from diamondback.interfaces.IProxy import IProxy
from diamondback.interfaces.IQ import IQ
from diamondback.interfaces.IRate import IRate
from diamondback.interfaces.IReset import IReset
from diamondback.interfaces.IResolution import IResolution
from diamondback.interfaces.IRotation import IRotation
from diamondback.interfaces.IS import IS
from diamondback.interfaces.IState import IState
from diamondback.interfaces.IStream import IStream
from diamondback.interfaces.ITimeZone import ITimeZone
from diamondback.interfaces.IUpdate import IUpdate
from diamondback.interfaces.IUrl import IUrl
from diamondback.interfaces.IUser import IUser
from diamondback.interfaces.IValid import IValid
from diamondback.interfaces.IVersion import IVersion
from diamondback.interfaces.IWrite import IWrite



from diamondback import clients
from diamondback import commons
from diamondback import filters
from diamondback import interfaces
from diamondback import models
from diamondback import transforms


__version__ = '1.0.35'


from diamondback.filters.FirFilter import FirFilter
from diamondback.filters.IirFilter import IirFilter
from diamondback.filters.ComplexBandPassFilter import ComplexBandPassFilter
from diamondback.filters.ComplexExponentialFilter import ComplexExponentialFilter
from diamondback.filters.ComplexFrequencyFilter import ComplexFrequencyFilter
from diamondback.filters.DerivativeFilter import DerivativeFilter
from diamondback.filters.GoertzelFilter import GoertzelFilter
from diamondback.filters.IntegralFilter import IntegralFilter
from diamondback.filters.PidFilter import PidFilter
from diamondback.filters.PolynomialRateFilter import PolynomialRateFilter
from diamondback.filters.PolyphaseRateFilter import PolyphaseRateFilter
from diamondback.filters.RankFilter import RankFilter
from diamondback.filters.WindowFilter import WindowFilter




""" **Description**

        Test models.

    **Example**

        ::

            pytest --verbose

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-04-03.

    **Definition**

"""

from diamondback.filters.ComplexExponentialFilter import ComplexExponentialFilter
from diamondback.filters.WindowFilter import WindowFilter
from diamondback.transforms.ComplexTransform import ComplexTransform
from diamondback.transforms.FourierTransform import FourierTransform
from diamondback.transforms.PowerSpectrumTransform import PowerSpectrumTransform
from diamondback.transforms.WaveletTransform import WaveletTransform
from diamondback.transforms.ZTransform import ZTransform
import math
import numpy


def test_ComplexTransform( count = 32, frequency = 0.1 ) :

    """ Test ComplexTransform.
    """

    x = ComplexExponentialFilter( 0.5 ).filter( numpy.linspace( -1.0e-4, 1.0e-4, count ) + frequency )

    for ii in range( 0, 2 ) :

        y = ComplexTransform.transform( x, ii == 0 )

        assert y.shape == ( 3, count )

        assert not isinstance( y[ 0 ], complex )

        z = ComplexTransform.transform( y, ii == 0 )

        assert z.shape == ( count, )

        assert isinstance( z[ 0 ], complex )

        assert numpy.allclose( z, x )


def test_FourierTransform( count = 128, frequency = ( 0.12, 0.23 ) ) :

    """ Test FourierTransform.
    """

    b = numpy.ones( count )

    for ii in range( 0, 4 ) :

        x = numpy.random.rand( 1 )[ 0 ] * ComplexExponentialFilter( numpy.random.rand( 1 )[ 0 ] ).filter( numpy.linspace( frequency[ 0 ], frequency[ 1 ], count ) )

        if ( ii % 2 ) :

            x = x.real

        if ( ii >= 2 ) :

            b = WindowFilter.Factory.instance( WindowFilter, 'Hann', count - 1 ).b

        y, f = FourierTransform.transform( x, b )

        assert ( ( len( y ) == count ) and ( len( f ) == count ) )

        assert isinstance( y[ 0 ], complex )

        assert numpy.isclose( f[ 0 ], -1.0 ) and numpy.isclose( f[ -1 ], 1.0 - 2.0 / count )

        h = numpy.argmax( abs( y[ count // 2 : ] ) )

        assert ( h >= numpy.argmin( abs( f[ count // 2 : ] - frequency[ 0 ] ) ) ) and ( h <= numpy.argmin( abs( f[ count // 2 : ] - frequency[ 1 ] ) ) )

        z = FourierTransform.transform( y, b, True )[ 0 ]

        if ( ii % 2 ) :

            z = z.real

        assert numpy.allclose( z[ 1 : -1 ], x[ 1 : -1 ] )


def test_PowerSpectrumTransform( count = 128, frequency = ( 0.12, 0.23 ) ) :

    """ Test PowerSpectrumTransform.
    """

    index = 64

    b = WindowFilter.Factory.instance( WindowFilter, 'Hann', count - 1 ).b

    for ii in range( 0, 2 ) :

        x = numpy.random.rand( 1 )[ 0 ] * ComplexExponentialFilter( numpy.random.rand( 1 )[ 0 ] ).filter( numpy.linspace( frequency[ 0 ], frequency[ 1 ], count * 8 ) )

        if ( ii > 0 ) :

            x = x.real

        y, f = PowerSpectrumTransform.transform( x, b, index )

        assert ( ( len( y ) == count ) and ( len( f ) == count ) )

        assert not isinstance( y[ 0 ], complex )

        assert numpy.isclose( f[ 0 ], -1.0 ) and numpy.isclose( f[ -1 ], 1.0 - 2.0 / count )

        h = numpy.argmax( y[ len( b ) // 2 : ] )

        assert ( h >= numpy.argmin( abs( f[ count // 2 : ] - frequency[ 0 ] ) ) ) and ( h <= numpy.argmin( abs( f[ count // 2 : ] - frequency[ 1 ] ) ) )


def test_WaveletTransform( count = 3 ) :

    """ Test WaveletTransform.
    """

    h = numpy.array( [  0.235233603892082,  0.570558457915722, 0.325182500263116, -0.095467207784164,
                       -0.060416104155198, 0.024908749868442 ] )

    g = numpy.array( [ -0.024908749868442, -0.060416104155198, 0.095467207784164,  0.325182500263116,
                       -0.570558457915722, 0.235233603892082 ] )

    obj = WaveletTransform.Factory.instance( WaveletTransform, 'Daubechies', len( h ) - 1 )

    assert numpy.allclose( obj.b[ 0 ][ 0 ], h )

    assert numpy.allclose( obj.b[ 0 ][ 1 ], g )

    h = numpy.array( [ 0.024908749868442, -0.060416104155198, -0.095467207784164, 0.325182500263116,
                       0.570558457915722,  0.235233603892082 ] )

    g = numpy.array( [ 0.235233603892082, -0.570558457915722,  0.325182500263116, 0.095467207784164,
                       -0.060416104155198, -0.024908749868442 ] )

    assert numpy.allclose( obj.b[ 1 ][ 0 ], h )

    assert numpy.allclose( obj.b[ 1 ][ 1 ], g )

    h = numpy.array( [ 0.5, 0.5 ] )

    g = numpy.array( [ 0.5, -0.5 ] )

    obj = WaveletTransform.Factory.instance( WaveletTransform, 'Haar', len( h ) - 1 )

    assert numpy.allclose( obj.b[ 0 ][ 0 ], h )

    assert numpy.allclose( obj.b[ 0 ][ 1 ], g )

    h = numpy.array( [ 0.5, 0.5 ] )

    g = numpy.array( [ -0.5, 0.5 ] )

    assert numpy.allclose( obj.b[ 1 ][ 0 ], h )

    assert numpy.allclose( obj.b[ 1 ][ 1 ], g )

    u = numpy.array( [ -0.677118658500953,  0.345009188833763,  0.118882064536892, -0.537364409907213,
                       -0.250663848240361,  0.491955501906881, -0.545336874797487,  0.390418096834100,
                       -0.271946313073118,  0.047745751406263,  0.301454810260592,  0.138563567406933,
                       -0.215817816000671, -0.271946313073118,  0.047745751406264,  0.301454810260592,
                       -0.154508497187474, -0.110615871041237, -0.024471741852423,  0.071019760960103,
                        0.139384128958763,  0.154508497187474,  0.110615871041237,  0.024471741852423,
                       -0.071019760960103, -0.139384128958763, -0.154508497187474, -0.110615871041237,
                       -0.024471741852423,  0.071019760960104,  0.139384128958763,  0.154508497187474 ] )

    x = ComplexExponentialFilter( 0.5 ).filter( numpy.ones( len( u ) ) * 0.1 ).real

    y = obj.transform( x, count, False )

    assert numpy.allclose( y, u )

    z = obj.transform( y, count, True )

    assert numpy.allclose( z, x )

    x = numpy.zeros( ( len( u ), len( u ) ) )

    for ii in range( 0, x.shape[ 0 ] ) :

        x[ ii, : ] = numpy.random.rand( 1 )[ 0 ] * ComplexExponentialFilter( numpy.random.rand( 1 )[ 0 ] * 2.0 - 1.0 ).filter( numpy.ones( len( u ) ) * numpy.random.rand( 1 )[ 0 ] * 0.5 ).real

    z = obj.transform( obj.transform( x, count, False ), count, True )

    assert numpy.allclose( z, x )


def test_ZTransform( ) :

    """ Test ZTransform.
    """

    frequency, ripple = 0.05, 0.125

    u = numpy.array( [  0.000000000000000, 1.868522837482727, -1.863713887857194, 0.948381518868702,
                       -0.212543935987133 ] )

    v = numpy.array( [ 0.016209591718306, 0.064838366873225, 0.097257550309837, 0.064838366873225,
                       0.01620959171830 ] )

    s = numpy.array( [ numpy.exp( 1j * math.pi * x / ( ( len( u ) - 1 ) * 2 ) ) for x in range( 1, ( len( u ) - 1 ) * 2, 2 ) ] )

    t = math.asinh( 1.0 / ( ( 10.0 ** ( 0.1 * ripple ) - 1.0 ) ** 0.5 ) ) / ( len( u ) - 1 )

    a = ( numpy.poly( ( -math.sinh( t ) * s.imag + 1j * math.cosh( t ) * s.real ) * 2.0 * math.pi ) ).real

    a /= a[ -1 ]

    b = [ 1.0 ]

    a, b = ZTransform.transform( a, b, frequency, True )

    b = numpy.poly( -numpy.ones( ( len( u ) - 1 ) ) )

    b *= ( 1.0 - sum( a ) ) / sum( b )

    assert numpy.allclose( a, u )

    assert numpy.allclose( b, v )

    frequency = 0.025

    u = numpy.array( [ 0.000000000000000, 4.913032434807922, -10.502699472790392, 12.647757190438988,
                       -9.256183138855487, 4.112750158020544, -1.0263940516247210, 0.110901278364195 ] )

    v = numpy.array( [ 0.000000000160721, 0.002515288055351, 0.103543670353901, 0.397913857193261,
                       0.290734727994417, 0.040371147586715, 0.000522947605366, 0.000000000000000 ] ) * 1.0e-3

    s, a = numpy.ones( 1 ), numpy.ones( 2 )

    for ii in range( 2, len( u ) ) :

        x = numpy.concatenate( ( s, numpy.zeros( 2 ) ) ) + numpy.concatenate( ( [ 0.0 ], ( ( 2.0 * ii ) - 1.0 ) * a ) )

        s, a = a, x

    a /= a[ -1 ]

    b = [ 1.0 ]

    a, b = ZTransform.transform( a, b, frequency, False )

    assert numpy.allclose( a, u )

    assert numpy.allclose( b, v )



""" **Description**

    REST client for simple REST service requests.  An API and an elective
    dictionary of parameter strings are encoded to build a URL, elective
    JSON or binary data are defined in the body of a request, and a JSON
    response is returned and decoded.

    A client instance may be useful as a base client definition to interact
    with a service which satisfies flexible request constraints.

    Caching may be useful in environments with intermittent or inconsistent
    network connectivity.  If caching is enabled, delete, patch, and put
    requests are cached and sent in order during a later request when the
    service is not ready, a property which may be overriden.

    URL and proxy definition is supported.

    Thread safe and reentrant.

    **Example**

        ::

            from diamondback.clients.RestClient import RestClient
            import requests
            import typing


            class TestClient( RestClient ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.cache = False

                    self.proxy = { 'http' : '', 'https' : '' }


                def add( self, item : typing.Dict[ str, float ] ) -> float :

                    return self.request( 'get', api = 'test/add', item = item )

            client = TestClient( )

            client.url = 'http://127.0.0.1:8080'

            client.add( { 'x', 2.71827, 'y' : 3.14159 } )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2020, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-10-22.

    **Definition**

"""

from diamondback.interfaces.ICache import ICache
from diamondback.interfaces.IData import IData
from diamondback.interfaces.IProxy import IProxy
from diamondback.interfaces.IUrl import IUrl
from threading import RLock
import requests
import typing


class RestClient( ICache, IData, IProxy, IUrl ) :

    """ REST client.
    """

    @property
    def live( self ) :

        """ Live ( bool ).
        """

        value = False

        try :

            value = self.request( 'get', 'live' )

        except :

            pass

        return value

    @property
    def ready( self ) :

        """ Ready ( bool ).
        """

        value = False

        try :

            value = ( ( self.live ) and ( self.request( 'get', 'ready' ) ) )

        except :

            pass

        return value

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._rlock = RLock( )

        self.cache, self.data = False, [ ]

        self.proxy, self.url = { }, 'http://127.0.0.1:8080'

    def request( self, method : str, api : str, item : typing.Dict[ str, str ] = None, json : any = None, data : any = None, timeout = 15.0 ) -> any :

        """ Request client for simple REST service requests. An API and an
            elective dictionary of parameter strings are encoded to build a
            URL, elective JSON or binary data are defined in the body of a
            request, and a JSON response is returned and decoded.

            Arguments :

                method - Method ( str ) in ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' ).

                api - API ( str ).

                item - Item ( dict( str, str ) ).

                json - JSON ( any ).

                data - Data ( any ).

                timeout - Timeout ( float ).

            Returns :

                value - Value ( any ).

        """

        if ( not method ) :

            raise ValueError( 'Method = ' + str( method ) )

        method = method.lower( )

        if ( method not in ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' ) ) :

            raise ValueError( 'Method = ' + str( method ) )

        if ( ( data ) and ( json ) ) :

            raise ValueError( '{:30s}{:30s}'.format( 'Data = ' + str( data ), 'Json = ' + str( json ) ) )

        if ( timeout <= 0.0 ) :

            raise ValueError( 'Timeout = ' + str( timeout ) )

        api = api.strip( '/' )

        url = self.url

        if ( api ) :

            url += '/' + api

        value = True

        with ( self._rlock ) :

            self.data.append( { 'method' : method, 'url' : url, 'item' : item, 'data' : data, 'json' : json } )

            if ( ( not self.cache ) or ( ( method not in ( 'delete', 'patch', 'put' ) ) or ( self.ready ) ) ) :

                for x in [ x for x in self.data ] :

                    try :

                        with requests.request( method = x[ 'method' ], url = x[ 'url' ], params = x[ 'item' ], data = x[ 'data' ], json = x[ 'json' ], proxies = self.proxy, timeout = timeout ) as value :

                            if ( ( not value ) or ( value.status_code != 200 ) ) :

                                raise ConnectionError( '{:30s}{:30s}'.format( 'Status = ' + str( value.status_code ), 'Reason = ' + str( value.reason ) ) )

                            value = value.json( )

                    finally :

                        del self.data[ 0 ]

        return value
